from __future__ import unicode_literals

from django.db import models
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from .compat import AUTH_USER_MODEL
from .settings import stored_messages_settings

INBOX_EXPIRE_DAYS = 30  # TODO move to settings


@python_2_unicode_compatible
class Message(models.Model):
    """
    TODO: docstring
    """
    message = models.TextField()
    level = models.IntegerField()
    tags = models.TextField()
    date = models.DateTimeField(default=timezone.now)

    def __str__(self):
        return self.message


@python_2_unicode_compatible
class MessageArchive(models.Model):
    """
    TODO: docstring
    """
    user = models.ForeignKey(AUTH_USER_MODEL)
    message = models.ForeignKey(Message)

    def __str__(self):
        return "[%s] %s" % (self.user, self.message)


@python_2_unicode_compatible
class Inbox(models.Model):
    """
    TODO: docstring
    """
    user = models.ForeignKey(AUTH_USER_MODEL)
    message = models.ForeignKey(Message)

    class Meta:
        verbose_name_plural = _('inboxes')

    def expired(self):
        expiration_date = self.message.date + timezone.timedelta(
            days=stored_messages_settings.INBOX_EXPIRE_DAYS)
        return expiration_date <= timezone.now()
    expired.boolean = True  # show a nifty icon in the admin

    def __str__(self):
        return "[%s] %s" % (self.user, self.message)



from rest_framework import viewsets
from rest_framework.decorators import action, api_view
from rest_framework.response import Response

from django.contrib.auth.decorators import login_required

from .models import Inbox
from .serializers import InboxSerializer


class InboxViewSet(viewsets.ReadOnlyModelViewSet):
    """
    Provides `list` and `detail` actions, plus a `read` POST endpoint for
    marking inbox messages as read.
    """
    serializer_class = InboxSerializer
    model = Inbox

    def get_queryset(self):
        if self.request.user.is_authenticated():
            return Inbox.objects.filter(user=self.request.user)
        return Inbox.objects.none()

    @action()
    def read(self, request, pk=None):
        """
        Mark the message as read (i.e. delete from inbox)
        """
        inbox_m = self.get_object()
        inbox_m.delete()
        return Response({'status': 'message marked as read'})


@login_required
@api_view(['POST'])
def mark_all_read(request):
    """
    TODO
    """
    Inbox.objects.filter(user=request.user).delete()
    return Response({"message": "All messages read"})



from rest_framework import serializers

from .models import Inbox


class InboxSerializer(serializers.ModelSerializer):
    class Meta:
        model = Inbox
        fields = ('id', 'message',)
        depth = 1



from django.contrib.messages.constants import *

STORED_DEBUG = DEBUG+1
STORED_INFO = INFO+1
STORED_SUCCESS = SUCCESS+1
STORED_WARNING = WARNING+1
STORED_ERROR = ERROR+1

DEFAULT_TAGS.update({
    STORED_DEBUG: 'persisted debug',
    STORED_INFO: 'persisted info',
    STORED_SUCCESS: 'persisted success',
    STORED_WARNING: 'persisted warning',
    STORED_ERROR: 'persisted error',
})



"""
This module is largely inspired by django-rest-framework settings.

Settings for the Stored Messages framework are all namespaced in the
STORED_MESSAGES setting.
For example your project's `settings.py` file might look like this::

    STORED_MESSAGES = {
        'INBOX_EXPIRE_DAYS': 30,
    }

This module provides the `stored_messages_settings` object, that is used to
access Stored Messages framework settings, checking for user settings first,
then falling back to the defaults.
"""
from __future__ import unicode_literals

from django.conf import settings
from django.utils import six, importlib

from .constants import *

USER_SETTINGS = getattr(settings, 'STORED_MESSAGES', None)
MESSAGE_TAGS = getattr(settings, 'MESSAGE_TAGS', {})

DEFAULTS = {
    'STORE_LEVELS': (
        STORED_DEBUG,
        STORED_INFO,
        STORED_SUCCESS,
        STORED_WARNING,
        STORED_ERROR,
    ),
    'MESSAGE_TAGS': MESSAGE_TAGS.update({
        STORED_DEBUG: 'stored debug',
        STORED_INFO: 'stored info',
        STORED_SUCCESS: 'stored success',
        STORED_WARNING: 'stored warning',
        STORED_ERROR: 'stored error',
    }),
    'INBOX_EXPIRE_DAYS': 30,
}

# List of settings that cannot be empty
MANDATORY = (

)

# List of settings that may be in string import notation.
IMPORT_STRINGS = (

)


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if isinstance(val, six.string_types):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        parts = val.split('.')
        module_path, class_name = '.'.join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except ImportError as e:
        msg = "Could not import '%s' for setting '%s'. %s: %s." % (val, setting_name,
                                                                   e.__class__.__name__, e)
        raise ImportError(msg)


class StoredMessagesSettings(object):
    """
    A settings object, that allows OAuth2 Provider settings to be accessed as properties.

    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """

    def __init__(self, user_settings=None, defaults=None, import_strings=None, mandatory=None):
        self.user_settings = user_settings or {}
        self.defaults = defaults or {}
        self.import_strings = import_strings or ()
        self.mandatory = mandatory or ()

    def __getattr__(self, attr):
        if attr not in self.defaults.keys():
            raise AttributeError("Invalid Stored Messages setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if val and attr in self.import_strings:
            val = perform_import(val, attr)

        self.validate_setting(attr, val)

        # Cache the result
        setattr(self, attr, val)
        return val

    def validate_setting(self, attr, val):
        if not val and attr in self.mandatory:
            raise AttributeError("Stored Messages setting: '%s' is mandatory" % attr)


stored_messages_settings = StoredMessagesSettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS,
                                                  MANDATORY)



"""
At the moment this module does something only when restframework is available
"""

try:
    from django.conf.urls import patterns, url, include
    from rest_framework.routers import DefaultRouter

    from . import views

    router = DefaultRouter()
    router.register(r'inbox', views.InboxViewSet)

    urlpatterns = patterns(
        '',
        url(r'^', include(router.urls)),
        url(r'^mark_all_read/$', views.mark_all_read),
    )
except ImportError:
    pass




__all__ = (
    'add_message_for', 'broadcast_message',
    'mark_read', 'mark_all_read',
)

from .models import Message, MessageArchive, Inbox


def add_message_for(users, level, message, extra_tags='', fail_silently=False):
    """
    Send a message to a list of users without passing through `django.contrib.messages`

    :param users: an iterable containing the recipients of the messages
    :param level: message level
    :param message: the string containing the message
    :param extra_tags: like the Django api, a string containing extra tags for the message
    :param fail_silently: not used at the moment
    """
    m = Message.objects.create(message=message, level=level, tags=extra_tags)
    for u in users:
        MessageArchive.objects.create(user=u, message=m)
        Inbox.objects.create(user=u, message=m)


def broadcast_message(level, message, extra_tags='', fail_silently=False):
    """
    Send a message to all users in the system. TODO.
    """
    # TODO
    raise NotImplementedError


def mark_read(user, message):
    """
    Mark message instance as read for user.
    Returns True if the message was `unread` and thus actually marked as `read` or False in case
    it is already `read` or it does not exist at all.

    :param user: user instance for the recipient
    :param message: a Message instance to mark as read
    """
    from .models import Inbox
    try:
        inbox_m = Inbox.objects.filter(user=user, message=message).get()
        inbox_m.delete()
        return True
    except Inbox.DoesNotExist:
        return False


def mark_all_read(user):
    """
    Mark all message instances for a user as read.

    :param user: user instance for the recipient
    """
    from .models import Inbox
    Inbox.objects.filter(user=user).delete()



from __future__ import unicode_literals

from django import template

from ..models import Inbox, MessageArchive


register = template.Library()


@register.inclusion_tag("stored_messages/stored_messages_list.html", takes_context=True)
def stored_messages_list(context, num_elements=10):
    """
    Renders a list of unread stored messages for the current user
    """
    if "user" in context:
        user = context["user"]
        if user.is_authenticated():
            return {
                "messages": Inbox.objects.select_related("message").filter(user=user)[:num_elements]
            }

@register.inclusion_tag("stored_messages/stored_messages_list.html", takes_context=True)
def stored_messages_archive(context, num_elements=10):
    """
    Renders a list of archived messages for the current user
    """
    if "user" in context:
        user = context["user"]
        if user.is_authenticated():
            return {
                "messages": MessageArchive.objects.select_related("message").filter(user=user)[:num_elements]
            }



from django.contrib.messages.storage.fallback import FallbackStorage

from .models import Inbox, Message, MessageArchive
from .settings import stored_messages_settings


class StorageMixin(object):
    """
    TODO: docstring
    """
    def __init__(self, request, *args, **kwargs):
        self.user = request.user
        super(StorageMixin, self).__init__(request, *args, **kwargs)

    def _get(self, *args, **kwargs):
        """
        Retrieve unread messages for current user, both from the inbox and
        from other storages
        """
        messages, all_retrieved = super(StorageMixin, self)._get(*args, **kwargs)
        if self.user.is_authenticated():
            inbox_messages = Inbox.objects.filter(user=self.user).select_related("message")
        else:
            inbox_messages = []
        return messages + [im.message for im in inbox_messages], all_retrieved

    def add(self, level, message, extra_tags=''):
        """
        If the message level was configured for being stored and request.user
        is not anonymous, save it to the database. Otherwise, let some other
        class handle the message.

        Notice: controls like checking the message is not empty and the level
        is above the filter need to be performed here, but it could happen
        they'll be performed again later if the message does not need to be
        stored.
        """
        if not message:
            return
        # Check that the message level is not less than the recording level.
        level = int(level)
        if level < self.level:
            return
        # Check if the message doesn't have a level that needs to be persisted
        if level not in stored_messages_settings.STORE_LEVELS or self.user.is_anonymous():
            return super(StorageMixin, self).add(level, message, extra_tags)

        self.added_new = True
        m = Message.objects.create(message=message, level=level, tags=extra_tags)
        MessageArchive.objects.create(user=self.user, message=m)
        self._queued_messages.append(m)

    def _store(self, messages, response, *args, **kwargs):
        """
        persistent messages are already in the database, so we can say they're
        already "stored"
        Here we put them in the inbox, or remove from the inbox in case the messages were
        iterated.

        messages contains only new msgs if self.used==True
        else contains both new and unread messages
        """
        contrib_messages = []
        if self.user.is_authenticated():
            if not messages:
                # erase inbox
                Inbox.objects.filter(user=self.user).delete()
            else:
                for m in messages:
                    if isinstance(m, Message):
                        # create inbox for the message if does not already exists
                        Inbox.objects.get_or_create(user=self.user, message=m)
                    else:
                        contrib_messages.append(m)

        super(StorageMixin, self)._store(contrib_messages, response, *args, **kwargs)

    def _prepare_messages(self, messages):
        """
        Like the base class method, prepares a list of messages for storage
        but avoid to do this for `models.Message` instances.
        """
        for message in messages:
            if not isinstance(message, Message):
                message._prepare()


class PersistentStorage(StorageMixin, FallbackStorage):
    """
    TODO: docstring
    """
    pass
